#include "UrsArduinoSoftI2cEx.h"

// Frequency  = 105 kHz, period = 9.5 us at a 16 MHz Arduino Uno

int8_t UrsArduinoSoftI2cExClass::begin(uint8_t sda, uint8_t scl) {
   // In order not to have to consider the current AVR type: get register pointers and bit mask with Arduino functions.
   sdaMask = digitalPinToBitMask(sda);
   sclMask = digitalPinToBitMask(scl);

   uint8_t port = digitalPinToPort(sda);
   if (port == NOT_A_PIN) return InvalidSda;
   sdaDDR = portModeRegister(port);
   sdaPIN = portInputRegister(port);
   volatile uint8_t* sdaPORT = portOutputRegister(port);

   port = digitalPinToPort(scl);
   if (port == NOT_A_PIN) return InvalidScl;
   sclDDR = portModeRegister(port);

   sclPIN = portInputRegister(port);

   volatile uint8_t* sclPORT = portOutputRegister(port);

   *sdaPORT &= ~(SoftI2cEx.sdaMask); // When the pin is switched to OUTPUT, it goes LOW.
   *sclPORT &= ~(SoftI2cEx.sclMask); // ... at INPUT, the external PULLUP pulls the level to HIGH.

   // Start with INPUT, i.e. both pins on HIGH.
   auto end = millis() + timeout;
   sdaHigh();
   while (!getSda() && (millis() < end)); // Able to control SDA?
   if (!getSda())
      return SdaError;

   return sclHigh();
}

int UrsArduinoSoftI2cExClass::writePacket(uint8_t deviceAdr, uint8_t* pData, uint8_t len, bool sendStop) {
   uint8_t count = 0;
   int8_t rtc;
   rtc = Start();
   if (rtc)
      goto Exit;
   rtc = WriteAddress((deviceAdr << 1) | 0x00);
   if (rtc) { // device address + write signal
      goto Exit;
   }

   for (uint8_t i = 0; i < len; i++) {
      if (rtc = Write(pData[i]))
         goto Exit;
      else
         count++;
   }
   rtc = count; // success

Exit:
   if (sendStop) {
      auto rtc1 = Stop();
      if (rtc1)
         rtc = rtc1; // Override count
   }
   return rtc;
}

int UrsArduinoSoftI2cExClass::readPacket(uint8_t deviceAdr, uint8_t* pData, uint8_t len) {
   int rtc;
   rtc = Start();
   if (rtc)
      goto Exit;

   rtc = WriteAddress((deviceAdr << 1) | 0x01);
   if (rtc) { // device address + read signal
      goto Exit;
   }

   for (uint8_t i = 0; i < len - 1; i++) { // Das letzte Byte wird separat eingelesen
      *pData++ = Read(true); // true: further bytes are requested
   }
   *pData = Read(false); // last Byte

   rtc = len;

Exit:
   auto rtc1 = Stop();
   if (rtc1)
      rtc = rtc1; // Override count

   return rtc;
}

int UrsArduinoSoftI2cExClass::writeToRegister(uint8_t deviceAdr, uint8_t regIndex, uint8_t* pData, uint8_t len) {
   uint8_t count = 0;
   int8_t rtc;
   rtc = Start();
   if (rtc)
      goto Exit;

   rtc = WriteAddress((deviceAdr << 1) | 0x00);
   if (rtc) { // device address + write signal
      goto Exit;
   }

   if (rtc = Write(regIndex))
      goto Exit;

   for (uint8_t i = 0; i < len; i++) {
      if (rtc = Write(pData[i]))
         goto Exit;
      else
         count++;
   }
   rtc = count; // success

Exit:
   auto rtc1 = Stop();
   if (rtc1)
      rtc = rtc1; // Override count

   return rtc;
}

int UrsArduinoSoftI2cExClass::readFromRegister(uint8_t deviceAdr, uint8_t regIndex, uint8_t* pData, uint8_t len, bool repeatedStart) {
   uint8_t rtc = writePacket(deviceAdr, &regIndex, 1, !repeatedStart); // Set Register index
   if (rtc != 1) {
      Stop();
      return rtc;
   }
   return readPacket(deviceAdr, pData, len);
}

int8_t UrsArduinoSoftI2cExClass::Start() {
   auto rtc = sclHigh();
   if (rtc)
      return rtc;
   sdaLow();
   return 0;
}

int8_t UrsArduinoSoftI2cExClass::Stop() {
   sdaLow();
   auto rtc = sclHigh();
   if (rtc)
      return rtc;
   sdaHigh();
   return 0;
}

int8_t UrsArduinoSoftI2cExClass::Write(uint8_t data) {
   int8_t rtc;

   for (uint8_t i = 0; i < 8; i++) {
      sclLow();

      if (data & 0x80)
         sdaHigh();
      else
         sdaLow();

      if (rtc = sclHigh())
         return rtc;

      data = data << 1;
   }

   //The 9th clock (ACK Phase)
   sclLow();

   sdaHigh();

   if (rtc = sclHigh())
      return rtc;

   int8_t ack = getSda() ? 1 : 0;

   sclLow();

   return ack;
}

int UrsArduinoSoftI2cExClass::Read(bool ack) {
   uint8_t data = 0x00;
   int8_t rtc;

   for (uint8_t i = 0; i < 8; i++) {
      sclLow();
      if (rtc = sclHigh())
         return rtc;
      data = data << 1;
      if (getSda())
         data |= 0x01;
   }

   sclLow();

   if (ack) {
      sdaLow();
   }
   else {
      sdaHigh();
   }

   if (rtc = sclHigh())
      return rtc;

   sclLow();
   sdaHigh(); // Release SDA

   return data;
}

UrsArduinoSoftI2cExClass SoftI2cEx;
